package org.icedrobot.daneel.dex;

import java.nio.BufferUnderflowException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;

public class Header {

  public enum Endianess { ENDIAN_CONSTANT, REVERSE_ENDIAN_CONSTANT }

  public static Header parse(ByteBuffer headerBuffer) {
    return new Header(headerBuffer);
  }

  private int checksum;

  private byte[] signature;

  private int fileLength;

  private int headerLength;

  private Endianess endianTag;

  private int linkSize;

  private int linkOff;

  private int mapOff;

  private int stringIdsSize;

  private int stringIdsOff;

  private int typeIdsSize;

  private int typeIdsOff;

  private int protoIdsSize;

  private int protoIdsOff;

  private int fieldIdsSize;

  private int fieldIdsOff;

  private int methodIdsSize;

  private int methodIdsOff;

  private int classDefsSize;

  private int classDefsOff;

  private int dataSize;

  private int dataOff;

  private Header(ByteBuffer buffer) {
    buffer = buffer.order(ByteOrder.LITTLE_ENDIAN);
    try {
      parseMagicByte(buffer);
      parseChecksum(buffer);
      parseSignature(buffer);
      parseFileLength(buffer);
      parseHeaderLength(buffer);
      parseEndianess(buffer);
      linkSize = buffer.getInt();
      linkOff = buffer.getInt();
      mapOff = buffer.getInt();
      stringIdsSize = buffer.getInt();
      stringIdsOff = buffer.getInt();
      typeIdsSize = buffer.getInt();
      typeIdsOff = buffer.getInt();
      protoIdsSize = buffer.getInt();
      protoIdsOff = buffer.getInt();
      fieldIdsSize = buffer.getInt();
      fieldIdsOff = buffer.getInt();
      methodIdsSize = buffer.getInt();
      methodIdsOff = buffer.getInt();
      classDefsSize = buffer.getInt();
      classDefsOff = buffer.getInt();
      dataSize = buffer.getInt();
      dataOff = buffer.getInt();
    } catch (BufferUnderflowException ex) {
      throw new DexParseException(ex);
    }
  }

  public int getChecksum() {
    return checksum;
  }

  public byte[] getSignature() {
    return signature;
  }

  public int getFileLength() {
    return fileLength;
  }

  public int getHeaderLength() {
    return headerLength;
  }

  public Endianess getEndianTag() {
    return endianTag;
  }

  public int getLinkSize() {
    return linkSize;
  }

  public int getLinkOff() {
    return linkOff;
  }

  public int getMapOff() {
    return mapOff;
  }

  public int getStringIdsSize() {
    return stringIdsSize;
  }

  public int getStringIdsOff() {
    return stringIdsOff;
  }

  public int getTypeIdsSize() {
    return typeIdsSize;
  }

  public int getTypeIdsOff() {
    return typeIdsOff;
  }

  public int getProtoIdsSize() {
    return protoIdsSize;
  }

  public int getProtoIdsOff() {
    return protoIdsOff;
  }

  public int getFieldIdsSize() {
    return fieldIdsSize;
  }

  public int getFieldIdsOff() {
    return fieldIdsOff;
  }

  public int getMethodIdsSize() {
    return methodIdsSize;
  }

  public int getMethodIdsOff() {
    return methodIdsOff;
  }

  public int getClassDefsSize() {
    return classDefsSize;
  }

  public int getClassDefsOff() {
    return classDefsOff;
  }

  public int getDataSize() {
    return dataSize;
  }

  public int getDataOff() {
    return dataOff;
  }

  private void parseMagicByte(ByteBuffer buffer) {
    assertMagicByte(buffer.get(), (byte) 0x64);
    assertMagicByte(buffer.get(), (byte) 0x65);
    assertMagicByte(buffer.get(), (byte) 0x78);
    assertMagicByte(buffer.get(), (byte) 0x0a);
    assertMagicByte(buffer.get(), (byte) 0x30);
    assertMagicByte(buffer.get(), (byte) 0x33);
    assertMagicByte(buffer.get(), (byte) 0x35);
    assertMagicByte(buffer.get(), (byte) 0x00);
  }

  private void assertMagicByte(byte actual, byte expected) {
    if (actual != expected) {
      throw new DexParseException();
    }
  }

  private void parseChecksum(ByteBuffer buffer) {
    checksum = buffer.getInt();
  }

  private void parseSignature(ByteBuffer buffer) {
    signature = new byte[20];
    buffer.get(signature, 0, 20);
  }

  private void parseFileLength(ByteBuffer buffer) {
    fileLength = buffer.getInt();
  }

  private void parseHeaderLength(ByteBuffer buffer) {
    headerLength = buffer.getInt();
  }

  private void parseEndianess(ByteBuffer buffer) {
    int endian = buffer.getInt();
    if (endian == 0x12345678) {
      endianTag = Endianess.ENDIAN_CONSTANT;
    } else if (endian == 0x78563412) {
      endianTag = Endianess.REVERSE_ENDIAN_CONSTANT;
    } else {
      throw new DexParseException("Invalid endian tag");
    }
  }
}
